@@ -1,4 +1,4 @@
-{-# LANGUAGE ScopedTypeVariables,DeriveDataTypeable #-}
+{-# LANGUAGE CPP,ScopedTypeVariables,DeriveDataTypeable #-}
 module Main where
 
 import Test.Framework (defaultMain, testGroup)
@@ -13,7 +13,9 @@ import Data.IORef
 import Data.Typeable
 import Control.Concurrent
 import Control.Monad
-import Data.List (sort)
+import Control.Applicative
+import Data.List (sort, permutations)
+import Data.Foldable (foldMap)
 import Data.Maybe
 
 import Prelude hiding (catch)
@@ -38,6 +40,8 @@ tests = [
   , testGroup "children surviving too long"
       [ testCase "concurrently+success" concurrently_success
       , testCase "concurrently+failure" concurrently_failure
+      , testCase "concurrentlyE+success" concurrentlyE_success
+      , testCase "concurrentlyE+failure" concurrentlyE_failure
       , testCase "race+success" race_success
       , testCase "race+failure" race_failure
       , testCase "cancel" cancel_survive
@@ -50,6 +54,18 @@ tests = [
   , testCase "link2" case_link2
   , testCase "link1_cancel" case_link1cancel
   , testCase "concurrently_deadlock" case_concurrently_deadlock
+  , testCase "concurrentlyE_deadlock" case_concurrentlyE_deadlock
+  , testGroup "concurrentlyE" [
+        testCase "concurrentlyE_right" concurrentlyE_right
+      , testCase "concurrentlyE_left1" concurrentlyE_left1
+      , testCase "concurrentlyE_left2" concurrentlyE_left2
+      , testCase "concurrentlyE_earlyException" concurrentlyE_earlyException
+      , testCase "concurrentlyE_lateException" concurrentlyE_lateException
+#if MIN_VERSION_base(4,9,0)
+      , testCase "concurrentlyE_Monoid" concurrentlyE_Monoid
+      , testCase "concurrentlyE_Monoid_fail" concurrentlyE_Monoid_fail
+#endif
+  ]
  ]
 
 value = 42 :: Int
@@ -173,6 +189,21 @@ concurrently_success = do
   res <- readIORef finalRes
   res @?= "parent"
 
+concurrentlyE_success :: Assertion
+concurrentlyE_success = do
+  finalRes <- newIORef "never filled"
+  baton <- newEmptyMVar
+  let quick = return (Right ())
+      slow = threadDelay 10000 *> return (Right ()) `finally` do
+        threadDelay 10000
+        writeIORef finalRes "slow"
+        putMVar baton ()
+  _ <- concurrentlyE quick slow
+  writeIORef finalRes "parent"
+  takeMVar baton
+  res <- readIORef finalRes
+  res @?= "parent"
+
 concurrently_failure :: Assertion
 concurrently_failure = do
   finalRes <- newIORef "never filled"
@@ -186,6 +217,19 @@ concurrently_failure = do
   res <- readIORef finalRes
   res @?= "parent"
 
+concurrentlyE_failure :: Assertion
+concurrentlyE_failure = do
+  finalRes <- newIORef "never filled"
+  let quick = error "a quick death"
+      slow = threadDelay 10000 *> return (Right ()) `finally` do
+        threadDelay 10000
+        writeIORef finalRes "slow"
+  _ :: Either SomeException (Either () ((), ())) <- try (concurrentlyE quick slow)
+  writeIORef finalRes "parent"
+  threadDelay 1000000 -- not using the baton, can lead to deadlock detection
+  res <- readIORef finalRes
+  res @?= "parent"
+
 race_success :: Assertion
 race_success = do
   finalRes <- newIORef "never filled"
@@ -348,3 +392,70 @@ case_concurrently_deadlock = do
     case e of
       Left BlockedIndefinitelyOnSTM{} -> True
       _other -> False
+
+-- See Issue #62
+case_concurrentlyE_deadlock :: Assertion
+case_concurrentlyE_deadlock = do
+  tvar <- newTVarIO False :: IO (TVar Bool)
+  e <- try $ void $ join (concurrentlyE) (fmap Right $ atomically $ readTVar tvar >>= check)
+    -- should throw BlockedIndefinitelyOnSTM not BlockedIndefinitelyOnMVar
+  assertBool "concurrentlyE_deadlock" $
+    case e of
+      Left BlockedIndefinitelyOnSTM{} -> True
+      _other -> False
+
+concurrentlyE_right :: Assertion
+concurrentlyE_right = do
+    r :: Either () (Bool,Bool) <- concurrentlyE (Right . const False <$> threadDelay 10000) (Right . const True <$> threadDelay 10000)
+    assertEqual "should be Right" (Right (False,True)) r
+
+concurrentlyE_left1 :: Assertion
+concurrentlyE_left1 = do
+    r :: Either () ((),()) <- concurrentlyE (Left <$> threadDelay 10000) (Right <$> forever (threadDelay 10000))
+    assertEqual "should be Left" (Left ()) r
+
+concurrentlyE_left2 :: Assertion
+concurrentlyE_left2 = do
+    r :: Either () ((),()) <- concurrentlyE (Right <$> forever (threadDelay 10000)) (Left <$> threadDelay 10000) 
+    assertEqual "should be Left" (Left ()) r
+
+concurrentlyE_earlyException :: Assertion
+concurrentlyE_earlyException = do
+    ref <- newIORef "never filled"
+    r :: Either TestException (Either () (Bool,Bool)) <- try $ 
+        concurrentlyE 
+            ((Right . const False <$> forever (threadDelay 10000)) `onException` writeIORef ref "finalized")
+            (threadDelay 1000 *> throwIO TestException)
+    refVal <- readIORef ref
+    assertEqual "should be Exception" (Left TestException, "finalized") (r, refVal)
+
+concurrentlyE_lateException :: Assertion
+concurrentlyE_lateException = do
+    ref <- newIORef "never filled"
+    r :: Either TestException (Either () (Bool,Bool)) <- try $ 
+        concurrentlyE 
+            ((Right . const False <$> threadDelay 100) `onException` writeIORef ref "finalized")
+            (threadDelay 100000 *> throwIO TestException)
+    refVal <- readIORef ref
+    assertEqual "should be Exception" (Left TestException, "never filled") (r, refVal)
+
+#if MIN_VERSION_base(4,9,0)
+concurrentlyE_Monoid :: Assertion
+concurrentlyE_Monoid = do
+    let delays :: [Int]
+        delays = [1000, 10000, 100000]
+        actions = zipWith (*>) (threadDelay <$> delays) (pure . Right . (:[]) <$> ['a'..])
+    r :: Either () String <- runConcurrentlyE $ foldMap ConcurrentlyE $ actions
+    assertEqual "Combined result in order" (Right "abc") r
+
+concurrentlyE_Monoid_fail :: Assertion
+concurrentlyE_Monoid_fail = do
+    let delays :: [Int]
+        delays = [1000, 200000]
+        actions = zipWith (*>) (threadDelay <$> delays) (pure . Right . (:[]) <$> ['a'..])
+        failDelays = [10000, 100000]
+        failActions = zipWith (*>) (threadDelay <$> delays) (pure . Left <$> ['u'..])
+    forM_ (permutations (actions ++ failActions)) $ \current -> do
+        r :: Either Char [Char] <- runConcurrentlyE $ foldMap ConcurrentlyE $ current
+        assertEqual "The earliest failure" (Left 'u') r
+#endif
